/*
 * mfe.c
 *
 * MPlayer frontend for the Nintendo GameCube 
 * Copyright (C) 2005 Albert Herranz
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 */

#include "config.h"
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <stdarg.h>
#include <errno.h>
#include <signal.h>
#include <string.h>
#include <math.h>
#include <dirent.h>
#include <mntent.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <linux/kd.h>

#include "SDL.h"
#include "SDL_ttf.h"
#include "SDL_keysym.h"

#define _GNU_SOURCE
#include <getopt.h>

#define SCREEN_WIDTH    640
#define SCREEN_HEIGHT   480
#define BPP             16

#define BUTTON_A        0
#define BUTTON_B        1
#define BUTTON_X        2
#define BUTTON_Y        3
#define BUTTON_Z        4
#define BUTTON_L        5
#define BUTTON_R        6
#define BUTTON_START    7

#define min(x,y) ({ \
	typeof(x) _x = (x);     \
	typeof(y) _y = (y);     \
	(void) (&_x == &_y);            \
	_x < _y ? _x : _y; })

#define PROC_MOUNTS "/proc/mounts"

#define MFE_VERSION "V" PACKAGE_VERSION "-01072005"

/*
 * Let's assume a 5% overscan on all sides of the screen.
 */
#define MFE_HORZ_OVERSCAN		((int)(SCREEN_WIDTH*0.05))
#define MFE_VERT_OVERSCAN		((int)(SCREEN_HEIGHT*0.05))
#define MFE_SCREEN_WIDTH		(SCREEN_WIDTH-2*MFE_HORZ_OVERSCAN)
#define MFE_SCREEN_HEIGHT		(SCREEN_HEIGHT-2*MFE_VERT_OVERSCAN)

/* font used in gui objects */
#define MFE_FONT			PKGDATADIR "/" "font.ttf"

/* basic artwork */
#define MFE_BACKGROUND			PKGDATADIR "/" "background.bmp"

#define MFE_ICON_ARROW_UP		PKGDATADIR "/" "up.bmp"
#define MFE_ICON_ARROW_DOWN		PKGDATADIR "/" "down.bmp"

/* text appended to scrolling labels */
#define MFE_SEPARATOR	"  *  "

/* kind of "typematic" stuff for joystick events */
#define MFE_DIALOG_JOYSTICK_REPEAT_DELAY	250 /* ms */
#define MFE_DIALOG_JOYSTICK_REPEAT_RATE		30 /* times per second */
#define MFE_PLAYER_JOYSTICK_REPEAT_DELAY	250 /* ms */
#define MFE_PLAYER_JOYSTICK_REPEAT_RATE		2 /* times per second */

/* time to wait for pclose to return before killing our child */
#define MFE_CHILD_STUCK_TIMEOUT		15

/* definitions related to dvd handling */
#define MFE_DVD_DEVICE			"/dev/dvd"
#define MFE_DVD_MOUNT_POINT		"/cdrom"

#define MFE_MPLAYER_EXTRA_OPTIONS	"-subfont-autoscale 1 -subpos 70" \
					" -sub-fuzziness 1"

SDL_Surface *screen;
SDL_Joystick *joystick;

SDL_Surface *background;
TTF_Font *font;
TTF_Font *font_small;

SDL_Surface *icon_arrow_up, *icon_arrow_down;

Uint32 cc_black, cc_white, cc_red;
Uint32 cc_bg, cc_fg;

enum {
	__STYLE_BOX = 0,
	__STYLE_BOX_COMPLETE,
	__STYLE_OSCILATE,
	__STYLE_SCROLL,
	__STYLE_TRANSITION,
};

enum {
	__ALIGN_CENTER = 0,
	__ALIGN_LEFT,
	__ALIGN_RIGHT,
};

enum {
	MFE_DOWN = 0,
	MFE_UP,
};

/*
 * A Label. Used to display text on the screen.
 */
struct mfe_label {
	int ticks;

	char *text;
	int len;

	int w;
	int h;

	TTF_Font *font;
	SDL_Color text_fg;
	SDL_Color text_shadow;
	SDL_Color box_bg;

	int style;
#define MFE_LABEL_STYLE_BOX		(1<<__STYLE_BOX)
#define MFE_LABEL_STYLE_BOX_COMPLETE	(1<<__STYLE_BOX_COMPLETE)
#define MFE_LABEL_STYLE_OSCILATE	(1<<__STYLE_OSCILATE)
#define MFE_LABEL_STYLE_SCROLL		(1<<__STYLE_SCROLL)
#define MFE_LABEL_STYLE_TRANSITION	(1<<__STYLE_TRANSITION)

	int h_align;
#define MFE_LABEL_ALIGN_CENTER		(1<<__ALIGN_CENTER)
#define MFE_LABEL_ALIGN_LEFT		(1<<__ALIGN_LEFT)
#define MFE_LABEL_ALIGN_RIGHT		(1<<__ALIGN_RIGHT)

	int max_w;
	int max_h;

	int needs_scrolling;
	int added_separator;
	int pos;
	int adjx;
	int first_glyph_advance;

	int throw;
	int spacing;

	char *next_text;
};

/*
 * An Item. Our basic element of a list box.
 */
struct mfe_item {
	struct mfe_item *next;
	struct mfe_label *label;

	int selected;

	void *private;
};

/*
 * A List Box. 
 */
struct mfe_list_box {
	struct mfe_item *items;
	int count;

	TTF_Font *font;
	int item_height;

	int saved_first_shown;
	int first_shown;
	int selected;

	int w;
	int h;

	int max_w;
	int max_h;

	int fancy;

	void *private;
};

/*
 * Our file dialog.
 */
struct mfe_file_dialog {
	struct mfe_list_box *lb;
	struct mfe_label *l;
	char *base;
	char *path;
	int action;
#define MFE_ACTION_NONE				0
#define MFE_ACTION_PLAY				1
#define MFE_ACTION_TOGGLE_MOUNTED_MEDIA		2
};

/*
 * A "Joystick Repeater". This is used to implement the joystick "typematic"
 * support.
 */
struct mfe_joystick_repeater {
	SDL_Joystick *joystick;

	int ticks;
	int repeating;

	int delay;
	int rate;

	int last_button_pressed;

	int last_hat_used;
	int last_state;

};

/*
 * A MPlayer session.
 */
struct mfe_session {
	FILE *mplayer;
	char *cmd;
	char *file;
	int flags;
#define MFE_SESSION_BROKEN	(1<<0)
#define MFE_SESSION_PAUSED	(1<<1)
};

/*
 * String handling functions.
 *
 *
 */

/*
 * Duplicates the text resulting from appending `a' to `b'.
 */
char *text_add_and_dup(char *a, char *b)
{
	char *s;

	s = (char *)malloc(strlen(a) + strlen(b) + 1);
	if (s) {
		strcpy(s, a);
		strcat(s, b);
	}
	return s;

}

/*
 * Duplicates a printf-like formatted string.
 */
static char *vstrdup(char *format, ...)
{
	va_list ap;
	int len, tentative_len;
	char *s;

	va_start(ap, format);

	tentative_len = strlen(format) + 1;
	s = (char *)malloc(tentative_len);
	if (s) {
		len = vsnprintf(s, tentative_len, format, ap);
		if (len >= tentative_len) {
			free(s);
			len++;
			s = (char *)malloc(len);
			if (s) {
				va_end(ap);
				va_start(ap, format);
				len = vsnprintf(s, len, format, ap);
				if (len < 0) {
					free(s);
					s = NULL;
				}
			}
		}
	}

	va_end(ap);

	return s;
}

static int ends_with(char *s1, char *s2)
{
	int len1, len2;

	len1 = strlen(s1);
	len2 = strlen(s2);

	if (len1 < len2)
		return 0;

	s1 += len1 - len2;
	return strcasecmp(s1, s2) == 0;
}


/*
 * File handling functions.
 *
 *
 */

/*
 * Removes the last part of a pathname, from the '/' until the end, except
 * for the root directory.
 */
static char *path_remove_last_name(char *path)
{
	char *s;

	s = strrchr(path, '/');
	if (s) {
		if (s == path) {
			s[0] = '/';
			s[1] = '\0';
		} else {
			s[0] = '\0';
		}
	}
	return path;
}

/*
 * Duplicates a path, adding a trailing '/' if necessary.
 */
char *path_dup_dirname(char *path)
{
	int len;
	char *s = path;

	len = strlen(s);
	if (s[len - 1] != '/') {
		path = (char *)malloc(len + 2);
		strcpy(path, s);
		path[len] = '/';
		path[len + 1] = '\0';
	} else {
		path = strdup(path);
	}
	return path;
}

/*
 * Duplicates a path resulting from appending `name' to `base'.
 */
char *path_dup_base_and_name(char *base, char *name)
{
	char *path;

	path = (char *)malloc(strlen(base) + strlen(name) + 1 + 1);
	if (path) {
		strcpy(path, base);
		strcat(path, "/");
		strcat(path, name);
	}
	return path;
}

/*
 * Compares two paths. Directory entries (ending in '/') have precedence.
 */
int path_compare(const void *a0, const void *b0)
{
	const char *a = a0, *b = b0;
	int lena, lenb;

	lena = strlen(a);
	lenb = strlen(b);

	if (a[lena - 1] == '/' && b[lenb - 1] != '/') {
		return -1;
	} else if (a[lena - 1] != '/' && b[lenb - 1] == '/') {
		return 1;
	}
	return strcmp(a, b);
}

/*
 * Gets the type of a file.
 */
static mode_t get_file_type(char *pathname)
{
	struct stat pstat;
	int retval;

	retval = lstat(pathname, &pstat);
	if (retval < 0) {
		return (mode_t) - 1;
	}
	return pstat.st_mode;
}

/*
 * Tells if a given path is a existing directory.
 */
static int is_a_dir(char *pathname)
{
	struct stat pstat;
	int retval;

	retval = stat(pathname, &pstat);
	if (retval < 0) {
		return 0;
	}
	return S_ISDIR(pstat.st_mode);
}

/*
 * Tells if the given path is currently a mount point for a filesystem.
 */
static int is_a_mount_point(char *path)
{
	FILE *f;
	struct mntent *entry;
	int found = 0;

	f = setmntent(PROC_MOUNTS, "r");
	if (f) {
		entry = getmntent(f);
		while (entry) {
			if (!strcmp(entry->mnt_dir, path)) {
				found = 1;
				break;
			}
			entry = getmntent(f);
		}
	}
	return found;
}

/*
 * Mounts the read only iso9660 filesystem found in `device' at `path'.
 */
static int mount_iso9660(char *device, char *path)
{
	char *cmd;
	int retval = -1;

	cmd = vstrdup("mount -t iso9660 -o ro %s %s", device, path);
	if (cmd) {
		retval = system(cmd);
		if (retval != 0) {
			fprintf(stderr, "mount command \"%s\" failed!\n", cmd);
		}
		free(cmd);
	}
	return retval;
}

/*
 * Unmounts the filesystem mounted at `path'.
 */
static int umount_iso9660(char *path)
{
	char *cmd;
	int retval = -1;

	cmd = vstrdup("umount %s", path);
	if (cmd) {
		retval = system(cmd);
		if (retval != 0) {
			fprintf(stderr, "umount command \"%s\" failed!\n", cmd);
		}
		free(cmd);
	}
	return retval;
}

/*
 * Drawing primitives.
 *
 *
 */

/*
 * Gets a color code given the current SDL color.
 */
Uint32 color_get_cc(SDL_PixelFormat * f, SDL_Color color)
{
	return SDL_MapRGBA(f, color.r, color.g, color.b, color.unused);
}

/*
 * Draws a pixel of color `color' at coordinates (`x', `y') of surface `s'.
 */
static void draw_pixel(SDL_Surface * s, Uint32 color, int x, int y)
{
	char *p = (char *)s->pixels;
	char *c = (char *)&color;

	if (x >= s->w || y >= s->h)
		return;

	if (SDL_BYTEORDER == SDL_BIG_ENDIAN)
		c += sizeof(Uint32) - s->format->BytesPerPixel;

	p += s->pitch * y;
	p += s->format->BytesPerPixel * x;

	memcpy(p, c, s->format->BytesPerPixel);
}

/*
 * Draws a horizontal line of length `len' starting at coordinates (`x',`y')
 * on surface `s'.
 */
static void draw_horz_line(SDL_Surface * s, Uint32 color, int x, int y, int len)
{
	char *p = (char *)s->pixels;
	char *c = (char *)&color;
	int bpp = s->format->BytesPerPixel;

	if (y >= s->h)
		return;

	if (SDL_BYTEORDER == SDL_BIG_ENDIAN)
		c += sizeof(Uint32) - s->format->BytesPerPixel;

	p += s->pitch * y;
	p += bpp * x;
	while (len > 0 && x < s->w) {
		memcpy(p, c, s->format->BytesPerPixel);
		p += bpp;
		x++;
		len--;
	}
}

/*
 * Draws a circle of color `color' and radius `radius' at
 * coordinates (`x',`y') of surface `s'.
 * The circle will be completely filled with `color` if `filled' is true.
 */
static void draw_circle(SDL_Surface * s, Uint32 color,
			int x, int y, int radius, int filled)
{
	long x_pos, y_pos, d, delta_e, delta_se;

	x_pos = -1;
	y_pos = radius;
	d = 1 - radius;
	delta_e = -1;
	delta_se = (-2 * radius) + 3;

	while (y_pos > x_pos) {
		delta_e += 2;
		x_pos++;

		if (d < 0) {
			d += delta_e;
			delta_se += 2;
		} else {
			d += delta_se;
			delta_se += 4;
			y_pos--;
		}

		if (filled) {
			draw_horz_line(s, color, x - x_pos, y - y_pos,
				       2 * x_pos);
			draw_horz_line(s, color, x - y_pos, y - x_pos,
				       2 * y_pos);
			draw_horz_line(s, color, x - y_pos, y + x_pos,
				       2 * y_pos);
			draw_horz_line(s, color, x - x_pos, y + y_pos,
				       2 * x_pos);
		} else {
			draw_pixel(s, color, x + x_pos, y - y_pos);
			draw_pixel(s, color, x - x_pos, y - y_pos);
			draw_pixel(s, color, x - y_pos, y - x_pos);
			draw_pixel(s, color, x + y_pos, y - x_pos);
			draw_pixel(s, color, x - y_pos, y + x_pos);
			draw_pixel(s, color, x + y_pos, y + x_pos);
			draw_pixel(s, color, x - x_pos, y + y_pos);
			draw_pixel(s, color, x + x_pos, y + y_pos);
		}
	}
}

/*
 * Draws a rectangle with rounded borders.
 * The rectangle will be transparent according to the passed alpha value,
 * and will let you see partially the specified background surface.
 */
static void draw_rounded_box(SDL_Surface * s, Uint32 color,
			     Uint8 alpha, SDL_Surface * s_bg,
			     int x0, int y0, int w, int h, int radius)
{
	SDL_Surface *box;
	SDL_PixelFormat *f;
	SDL_Rect r0, r;
	int x = 0, y = 0;

	if (w <= 0 || h <= 0)
		return;

	if (x0 < 0)
		x0 = 0;
	if (y0 < 0)
		y0 = 0;
	if (w > s->w)
		w = s->w;
	if (h > s->h)
		h = s->h;

	radius = min(radius, h / 2);
	radius = min(radius, w / 2);

	if (s_bg) {
		f = s_bg->format;
	} else {
		f = screen->format;
	}

	box = SDL_CreateRGBSurface(SDL_SWSURFACE, w, h, f->BitsPerPixel,
				   f->Rmask, f->Gmask, f->Bmask, f->Amask);
	SDL_SetAlpha(box, SDL_SRCALPHA, alpha);

	r0.x = x0;
	r0.y = y0;
	r0.w = w;
	r0.h = h;
	if (s_bg) {
		r.x = x;
		r.y = y;
		SDL_BlitSurface(s_bg, &r0, box, &r);
	} else {
		SDL_FillRect(box, NULL, cc_bg);
	}

	if (radius >= 1) {
		draw_circle(box, color, x + radius, y + radius, radius, 1);
		draw_circle(box, color, x + w - radius, y + radius, radius, 1);
		draw_circle(box, color, x + radius, y + h - radius, radius, 1);
		draw_circle(box, color, x + w - radius, y + h - radius,
			    radius, 1);
		r.x = x + radius;
		r.y = y;
		r.w = w - 2 * radius;
		r.h = h;
		SDL_FillRect(box, &r, color);
	}
	r.x = x;
	r.y = y + radius;
	r.w = w;
	r.h = h - 2 * radius;
	SDL_FillRect(box, &r, color);

	SDL_BlitSurface(box, NULL, s, &r0);
	SDL_FreeSurface(box);
}

/*
 * Draws the image `image' at coordinates (`x',`y') of surface `s'.
 */
static void draw_image(SDL_Surface * s, int x, int y, SDL_Surface * image)
{
	SDL_Rect dest = { x, y, 0, 0 };

	SDL_BlitSurface(image, NULL, s, &dest);
}

/*
 * GUI handling routines.
 *
 *
 */

/*
 * Loads a (BMP) image from a file.
 */
static SDL_Surface *mfe_load_image(char *file)
{
	SDL_Surface *image = NULL, *temp;

	temp = SDL_LoadBMP(file);
	if (temp) {
		image = SDL_DisplayFormat(temp);
		SDL_FreeSurface(temp);
	}
	return image;
}

/*
 * Loads a (TTF) font from a file.
 */
static TTF_Font *mfe_load_font(char *file, int ptsize)
{
	TTF_Font *font;

	font = TTF_OpenFont(file, ptsize);
	if (!font) {
		fprintf(stderr, "Unable to load font %s: %s.\n",
			file, TTF_GetError());
	}
	return font;
}

/*
 * Leaves graphics mode.
 */
static void mfe_leave_graphics(void)
{
	SDL_FreeSurface(background);
	background = NULL;

	TTF_CloseFont(font);
	font = NULL;

	SDL_QuitSubSystem(SDL_INIT_VIDEO);
}

/*
 * Enters graphics mode.
 */
static int mfe_enter_graphics(int flags)
{
	SDL_InitSubSystem(SDL_INIT_VIDEO);

	screen = SDL_SetVideoMode(SCREEN_WIDTH, SCREEN_HEIGHT, BPP, flags);
	if (screen == NULL) {
		fprintf(stderr, "Couldn't set %dx%d video mode: %s.",
			SCREEN_WIDTH, SCREEN_HEIGHT, SDL_GetError());
		return -1;
	}
	SDL_ShowCursor(SDL_DISABLE);

	cc_red = SDL_MapRGB(screen->format, 255, 0, 0);
	cc_white = SDL_MapRGB(screen->format, 255, 255, 255);
	cc_black = SDL_MapRGB(screen->format, 0, 0, 0);

	cc_bg = cc_white;
	cc_fg = cc_black;

	background = mfe_load_image(MFE_BACKGROUND);

	font = mfe_load_font(MFE_FONT, 24);
	font_small = mfe_load_font(MFE_FONT, 18);
	//font_big = mfe_load_font(MFE_FONT, 42);

	icon_arrow_down = mfe_load_image(MFE_ICON_ARROW_DOWN);
	SDL_SetColorKey(icon_arrow_down, SDL_SRCCOLORKEY, cc_black);
	icon_arrow_up = mfe_load_image(MFE_ICON_ARROW_UP);
	SDL_SetColorKey(icon_arrow_up, SDL_SRCCOLORKEY, cc_black);

	return 0;
}

/*
 * Exits the grapical user interface.
 */
static void mfe_exit_gui(void)
{
	mfe_leave_graphics();

	if (joystick)
		SDL_JoystickClose(joystick);
	SDL_Quit();
}

/*
 * Initializes the graphical user interface.
 */
static int mfe_init_gui(int flags)
{
	if (SDL_Init(SDL_INIT_JOYSTICK) < 0) {
		fprintf(stderr, "Couldn't initialize SDL: %s.\n",
			SDL_GetError());
		return -1;
	}
	atexit(mfe_exit_gui);

	if (TTF_Init() < 0) {
		fprintf(stderr, "Couldn't initialize SDL_ttf: %s.",
			TTF_GetError());
		return -1;
	}
	atexit(TTF_Quit);

	if (SDL_NumJoysticks() > 0) {
		SDL_JoystickEventState(SDL_ENABLE);
		joystick = SDL_JoystickOpen(0);
	}

	mfe_enter_graphics(flags);

	return 0;
}

/*
 * Draws the specified image as a background image.
 */
static void mfe_draw_background(SDL_Surface * image)
{
	SDL_Rect dest = { (SCREEN_WIDTH - image->w) / 2,
		(SCREEN_HEIGHT - image->h) / 2,
		0, 0
	};

	if (!image || !screen)
		return;

	SDL_FillRect(screen, NULL, cc_bg);
	draw_image(screen, dest.x, dest.y, image);
	SDL_UpdateRect(screen, 0, 0, 0, 0);
}

/*
 * Our GUI components: labels, items and list boxes.
 *
 *
 */

#define AMPL_X 4.0
#define AMPL_Y 12.0
#define AMPL_TIME 16.0

static void mfe_image_draw(SDL_Rect * dest, SDL_Surface * image);
static void mfe_label_recalc_rendered_size(struct mfe_label *l);
static void mfe_label_draw_blank(SDL_Rect * dest, struct mfe_label *l);

/*
 * Draws a blank image at `rect'.
 * A blank image in fact exposes the background image at `rect'.
 */
static void mfe_image_draw_blank(SDL_Rect * dest)
{
	if (background) {
		SDL_BlitSurface(background, dest, screen, dest);
	} else {
		SDL_FillRect(screen, dest, cc_bg);
	}
	SDL_UpdateRect(screen, dest->x, dest->y, dest->w, dest->h);
}

/*
 * Draws the image `image' at `rect'.
 */
static void mfe_image_draw(SDL_Rect * dest, SDL_Surface * image)
{
	if (image) {
		SDL_BlitSurface(image, NULL, screen, dest);
		SDL_UpdateRect(screen, dest->x, dest->y, dest->w, dest->h);
	} else {
		mfe_image_draw_blank(dest);
	}
}

/*
 * Initializes a label with default values.
 */
static void mfe_label_init(struct mfe_label *l)
{
	SDL_Color text_fg = { 200, 200, 200, 0 };
	SDL_Color text_shadow = { 0, 0, 0, 255 };
	SDL_Color box_bg = { 0, 0, 0, 255 };

	memset(l, 0, sizeof(*l));

	l->style = MFE_LABEL_STYLE_SCROLL | MFE_LABEL_STYLE_TRANSITION;
	l->h_align = MFE_LABEL_ALIGN_CENTER;

	l->max_w = SCREEN_WIDTH;
	l->max_h = SCREEN_HEIGHT;

	l->font = font;
	l->text_fg = text_fg;
	l->text_shadow = text_shadow;
	l->box_bg = box_bg;
}

/*
 * Creates and initializes a label.
 */
static struct mfe_label *mfe_label_create(void)
{
	struct mfe_label *l;

	l = (struct mfe_label *)malloc(sizeof(*l));
	if (l) {
		mfe_label_init(l);
	}
	return l;
}

/*
 * Frees the texts stored in a label.
 */
static void mfe_label_free_text(struct mfe_label *l)
{
	if (l->next_text) {
		free(l->next_text);
		l->next_text = NULL;
	}
	if (l->text) {
		free(l->text);
		l->text = NULL;
		l->len = 0;
	}
}

/*
 * Destroys a label.
 */
void mfe_label_destroy(struct mfe_label *l)
{
	mfe_label_free_text(l);
	free(l);
}

/*
 * Sets the font of a label.
 */
static inline void mfe_label_set_font(struct mfe_label *l, TTF_Font * f)
{
	l->font = f;
}

/*
 * Sets the foreground color of a label.
 */
static inline void mfe_label_set_color_fg(struct mfe_label *l, SDL_Color fg)
{
	l->text_fg = fg;
}

/*
 * Sets the color used to draw the bounding box of a "boxed" label.
 */
static inline void mfe_label_set_color_bg(struct mfe_label *l, SDL_Color bg)
{
	l->box_bg = bg;
}

/*
 * Cancels a "throw" effect.
 */
static void label_cancel_throw(struct mfe_label *l)
{
	l->throw = 0;
	l->spacing = 0;
}

/*
 * Starts a "left throw" effect.
 */
static void label_throw_left(struct mfe_label *l)
{
	l->throw = -1;
	l->spacing = l->w;
}

/*
 * Starts a "right throw" effect.
 */
static void label_throw_right(struct mfe_label *l)
{
	l->throw = 1;
	l->spacing = 1;
}

/*
 * Switches the current text of a label with the previously loaded
 * pending text, using a visual effect.
 */
static void label_switch_text(struct mfe_label *l)
{
	if (l->text)
		free(l->text);
	l->text = l->next_text;
	l->next_text = NULL;

	if (l->text) {
		l->pos = 0;
		l->added_separator = 0;
		label_throw_left(l);
	}
}

/*
 * Sets the text of a label, using a visual effect.
 */
static char *mfe_label_set_text(struct mfe_label *l, char *text)
{
	if (l->next_text) {
		free(l->next_text);
		l->next_text = NULL;
	}

	if (text) {
		l->next_text = strdup(text);
		if (!l->next_text)
			return NULL;

		if (l->text) {
			label_throw_right(l);
		} else {
			label_switch_text(l);
		}
	} else {
		label_throw_right(l);
	}

	return l->text;
}

/*
 * Add the given styles to the label.
 */
static int mfe_label_add_styles(struct mfe_label *l, int styles)
{
	l->style |= styles;
	return l->style;
}

/*
 * Remove the given styles from the label.
 */
static int mfe_label_remove_styles(struct mfe_label *l, int styles)
{
	l->style &= ~styles;
	return l->style;
}

/*
 * Handle a part of a "throw" effect.
 */
static int label_pre_handle_throw(SDL_Rect * dest, struct mfe_label *l)
{
	int transition = (l->style & MFE_LABEL_STYLE_TRANSITION);

	if (l->throw > 0 && (l->spacing >= l->max_w || !transition)) {
		/* right throw */
		if (l->next_text) {
			label_switch_text(l);
		} else {
			/* event here */
			mfe_label_draw_blank(dest, l);
			mfe_label_free_text(l);
			return 1;
		}
	} else if (l->throw < 0 && (l->spacing <= 0 || !transition)) {
		/* left throw */
		/* event here */
		mfe_label_draw_blank(dest, l);
		//return 1;
	}

	return 0;
}

/*
 * Handle the remaining part of a "throw" effect.
 */
static void label_post_handle_throw(struct mfe_label *l)
{
	char *c;

	if (l->throw) {
		l->spacing += l->throw;
		if (l->throw > 0)
			l->throw++;
		else
			l->throw--;
	} else {
		if (l->needs_scrolling) {
			l->adjx--;
			if (l->adjx <= 0) {
				c = l->text + l->pos;
				l->pos++;
				l->adjx = l->first_glyph_advance;
			}
		}
	}
}

/*
 * Returns the width of a label.
 */
static inline int mfe_label_get_width(struct mfe_label *l)
{
	return l->w;
}

/*
 * Returns the height of a label.
 */
static inline int mfe_label_get_height(struct mfe_label *l)
{
	return l->h;
}

/*
 * Clips a label into the given width and height.
 */
static void mfe_label_clip(struct mfe_label *l, int w, int h)
{
	l->max_w = w;
	l->max_h = h;
}

/*
 * Calculates the rendered size of a label.
 */
static void mfe_label_recalc_rendered_size(struct mfe_label *l)
{
	int minx, miny, maxy, advance;
	int width;

	char *c, *temp;
	int len;

	SDL_Surface *glyph;

	if (!l->text)
		return;

	len = strlen(l->text);
	l->len = len;

	if (l->pos >= len)
		l->pos = 0;
	c = l->text + l->pos;

	width = 4;
	if ((l->style & MFE_LABEL_STYLE_OSCILATE))
		width += AMPL_X + AMPL_X;

	l->first_glyph_advance = -1;

	while (len > 0) {
		width += l->spacing;

		if (*c == 0) {
			c = l->text;
		}

		TTF_GlyphMetrics(l->font, *c,
				 &minx, NULL, &miny, &maxy, &advance);
		glyph = TTF_RenderGlyph_Blended(l->font, *c, l->text_fg);
		if (*c == ' ') {
			advance = minx + advance;
		} else {
			advance = minx + glyph->w;
		}
		SDL_FreeSurface(glyph);

		if (l->first_glyph_advance == -1) {
			l->first_glyph_advance = advance;
			if ((l->style & MFE_LABEL_STYLE_OSCILATE))
				l->first_glyph_advance += AMPL_X;
		}

		width += advance;
		if ((l->style & MFE_LABEL_STYLE_OSCILATE))
			width += AMPL_X;
		if (width > l->max_w) {
			width = l->max_w;
			break;
		}

		c++;
		len--;
	}

	l->w = width;
	l->h = TTF_FontHeight(l->font) + 4;
	if ((l->style & MFE_LABEL_STYLE_OSCILATE))
		l->h += 2 * AMPL_Y;
	l->h = min(l->h, l->max_h);

	l->needs_scrolling = 0;
	if (len > 0 && !l->throw) {
		if ((l->style & MFE_LABEL_STYLE_SCROLL)) {
			l->needs_scrolling = 1;
			if (!l->added_separator) {
				temp = text_add_and_dup(l->text, MFE_SEPARATOR);
				if (temp) {
					free(l->text);
					l->text = temp;
					l->added_separator = 1;
				}
			}
		}
	}
}

/*
 * Draws a blank label.
 * (A blank label actually exposes the background occupied by the label).
 */
static void mfe_label_draw_blank(SDL_Rect * dest, struct mfe_label *l)
{
	SDL_Rect r;

	mfe_label_clip(l, dest->w, dest->h);
	//mfe_label_recalc_rendered_size(l);

	r.x = dest->x;
	r.y = dest->y;
	r.w = dest->w;
	r.h = dest->h;

	mfe_image_draw_blank(&r);

	label_cancel_throw(l);
}

/*
 * Prepares a bit the space where the label will be rendered.
 * (It exposes the background and draws a box as needed).
 */
static void label_draw_prepare(SDL_Rect * r, SDL_Rect * dest,
			       struct mfe_label *l)
{
	int x;
	Uint32 cc = color_get_cc(screen->format, l->box_bg);

	if (l->h_align == MFE_LABEL_ALIGN_CENTER)
		x = (dest->w - l->w) / 2;
	else
		x = 0;

	r->x = dest->x;
	r->w = dest->w;
	r->y = dest->y;
	r->h = l->h;

	if (background) {
		SDL_BlitSurface(background, r, screen, r);
	} else {
		SDL_FillRect(screen, r, cc_bg);
	}

	if (!(l->style & MFE_LABEL_STYLE_BOX_COMPLETE)) {
		r->x = dest->x + x;
		r->w = l->w;
	}
#define MFE_LABEL_BOX_ALPHA 64

	if ((l->style & MFE_LABEL_STYLE_BOX) ||
	    (l->style & MFE_LABEL_STYLE_BOX_COMPLETE)) {
		draw_rounded_box(screen, cc, MFE_LABEL_BOX_ALPHA,
				 background,
				 r->x, r->y, r->w, r->h,
				 TTF_FontHeight(l->font) / 2);
	}

	if ((l->style & MFE_LABEL_STYLE_BOX_COMPLETE)) {
		r->x += x;
	}
}

/*
 * Renders the text of a label.
 */
static void label_draw_continue(SDL_Rect * r, SDL_Rect * dest,
				struct mfe_label *l)
{
	int minx, miny, maxy, advance;
	int x, y;
	double x_delta, y_delta;
	char *c;
	int len;
	int ticks;

	SDL_Surface *glyph;

	ticks = l->ticks;

	x_delta = 0;
	y_delta = 0;

	x = r->x;
	y = r->y;

	if ((l->style & MFE_LABEL_STYLE_OSCILATE)) {
		x += AMPL_X;
		y += AMPL_Y;
		r->w -= 2 * AMPL_X;
	}

	if (l->needs_scrolling) {
		x += l->adjx;
	}

	len = l->len;
	c = l->text + l->pos;
	while (len > 0) {
		if (*c == 0) {
			c = l->text;
		}

		TTF_GlyphMetrics(l->font, *c,
				 &minx, NULL, &miny, &maxy, &advance);
		glyph = TTF_RenderGlyph_Blended(l->font, *c, l->text_shadow);

		if ((l->style & MFE_LABEL_STYLE_OSCILATE)) {
			x_delta = AMPL_X * cos(ticks / AMPL_TIME);
			y_delta = AMPL_Y * sin(ticks / AMPL_TIME);
		}

		x += l->spacing;

		r->w = 0;
		r->h = 0;

		r->x = x + minx + x_delta;
		r->y = y + TTF_FontAscent(l->font) - maxy + y_delta;

		if (r->x + glyph->w > dest->x + dest->w) {
			SDL_FreeSurface(glyph);
			break;
		}

		SDL_SetAlpha(glyph, SDL_SRCALPHA, 128);
		SDL_BlitSurface(glyph, NULL, screen, r);
		r->x += 4;
		r->y += 4;
		SDL_BlitSurface(glyph, NULL, screen, r);
		SDL_FreeSurface(glyph);
		glyph = TTF_RenderGlyph_Blended(l->font, *c, l->text_fg);
		r->x -= 2;
		r->y -= 2;
		SDL_BlitSurface(glyph, NULL, screen, r);

		if (*c == ' ') {
			x += minx + advance;
		} else {
			x += minx + glyph->w;
		}

		if ((l->style & MFE_LABEL_STYLE_OSCILATE)) {
			x += AMPL_X;
		}

		SDL_FreeSurface(glyph);

		ticks++;

		c++;
		len--;
	}
}

/*
 * Starts a "fancy" transition for a label.
 * This actually makes the label unuseable, so the label must be destroyed
 * after the fancy transition finishes.
 */
static void mfe_label_start_fancy_transition(struct mfe_label *l)
{
	mfe_label_set_text(l, NULL);
}

/*
 * Checks if a "fancy" transition has finished for the given label.
 */
static int mfe_label_check_for_fancy_transition(struct mfe_label *l)
{
	return l->text != NULL;
}

/*
 * Draws a label to the screen at the given rectangle.
 */
static void mfe_label_draw(SDL_Rect * dest, struct mfe_label *l)
{
	SDL_Rect r;

	mfe_label_clip(l, dest->w, dest->h);

	if (!l->text) {
		mfe_label_draw_blank(dest, l);
		return;
	}

	if (label_pre_handle_throw(dest, l))
		return;

	l->ticks++;

	mfe_label_recalc_rendered_size(l);

	label_draw_prepare(&r, dest, l);
	label_draw_continue(&r, dest, l);

	SDL_UpdateRect(screen, dest->x, dest->y, dest->w, l->h);

	label_post_handle_throw(l);
}

/*
 * Initializes an item with default values.
 */
static void mfe_item_init(struct mfe_item *i)
{
	memset(i, 0, sizeof(*i));
}

/*
 * Creates and initializes a new item.
 */
static struct mfe_item *mfe_item_create(void)
{
	struct mfe_item *i;

	i = (struct mfe_item *)malloc(sizeof(*i));
	if (i) {
		mfe_item_init(i);
		i->label = mfe_label_create();
		if (i->label)
			mfe_label_init(i->label);
	}
	return i;
}

/*
 * Destroys an item.
 */
static void mfe_item_destroy(struct mfe_item *i)
{
	if (i->label)
		mfe_label_destroy(i->label);
	i->label = NULL;
}

/*
 * Sets the text of an item.
 */
static void mfe_item_set_text(struct mfe_item *i, char *text)
{
	if (i->label)
		mfe_label_set_text(i->label, text);
}

/*
 * Tells if the given item is selected.
 */
static inline int mfe_item_is_selected(struct mfe_item *i)
{
	return i->selected;
}

/*
 * Selects or deselects the given item, according to `select'.
 */
static int mfe_item_select(struct mfe_item *i, int select)
{
	if (!i->label)
		return -1;

	if (select) {
		mfe_label_add_styles(i->label, MFE_LABEL_STYLE_BOX_COMPLETE |
				     MFE_LABEL_STYLE_SCROLL);
		i->selected = 1;
	} else {
		mfe_label_remove_styles(i->label, MFE_LABEL_STYLE_BOX_COMPLETE |
					MFE_LABEL_STYLE_SCROLL);
		i->selected = 0;
	}
	return i->selected;
}

/*
 * Returns the width of the given item.
 */
static inline int mfe_item_get_width(struct mfe_item *i)
{
	if (i->label)
		return mfe_label_get_width(i->label);
	return 0;
}

/*
 * Returns the height of the given item.
 */
static inline int mfe_item_get_height(struct mfe_item *i)
{
	if (i->label)
		return mfe_label_get_height(i->label);
	return 0;
}

/*
 * Clips an item into the given width and height.
 */
static void mfe_item_clip(struct mfe_item *i, int w, int h)
{
	if (i->label)
		mfe_label_clip(i->label, w, h);
}

/*
 * Calculates the rendered size of an item.
 */
static void mfe_item_recalc_rendered_size(struct mfe_item *i)
{
	if (i->label)
		mfe_label_recalc_rendered_size(i->label);
}

/*
 * Draws an item.
 */
static void mfe_item_draw(SDL_Rect * dest, struct mfe_item *i)
{
	mfe_item_clip(i, dest->w, dest->h);
	if (i->label)
		mfe_label_draw(dest, i->label);
}

static void mfe_list_box_draw(SDL_Rect * dest, struct mfe_list_box *lb);

/*
 * Initializes a list box with default values.
 */
void mfe_list_box_init(struct mfe_list_box *lb)
{
	memset(lb, 0, sizeof(*lb));

	lb->font = font;
	lb->item_height = TTF_FontHeight(lb->font) + 4;

	lb->fancy = -1;

	lb->max_w = SCREEN_WIDTH;
	lb->max_h = SCREEN_HEIGHT;
}

/*
 * Creates and initializes a list box.
 */
struct mfe_list_box *mfe_list_box_create(void)
{
	struct mfe_list_box *lb;

	lb = (struct mfe_list_box *)malloc(sizeof(*lb));
	if (lb) {
		mfe_list_box_init(lb);
	}
	return lb;
}

/*
 * Adds an item to a list box, at the beginning.
 */
void mfe_list_box_link_item(struct mfe_list_box *lb, struct mfe_item *i)
{
	if (lb->items) {
		i->next = lb->items;
	}
	lb->items = i;
	lb->count++;
}

/*
 * Removes an item from a list box.
 */
void mfe_list_box_unlink_item(struct mfe_list_box *lb, struct mfe_item *i)
{
	struct mfe_item *iter = lb->items;

	if (!i)
		return;

	while (iter) {
		if (iter == i) {
			iter->next = i->next;
			lb->count--;
			break;
		}
		iter = iter->next;
	}

}

/*
 * Returns the item at the given position in a list box.
 */
struct mfe_item *mfe_list_box_get_item_by_index(struct mfe_list_box *lb,
						int index)
{
	struct mfe_item *item = lb->items;
	int i = 0;

	if (index >= 0) {
		while (item) {
			if (i == index) {
				return item;
			}
			item = item->next;
			i++;
		}
	}
	return NULL;
}

/*
 * Returns the selected item of a list box.
 */
struct mfe_item *mfe_list_box_get_selection(struct mfe_list_box *lb)
{
	return mfe_list_box_get_item_by_index(lb, lb->selected);
}

/*
 * Destroys the given item, unlinking it from the list box.
 */
void mfe_list_box_destroy_item(struct mfe_list_box *lb, struct mfe_item *i)
{
	mfe_list_box_unlink_item(lb, i);
	mfe_item_destroy(i);
}

/*
 * Destroys all items in a list box.
 */
void mfe_list_box_destroy_all_items(struct mfe_list_box *lb)
{
	struct mfe_item *temp;

	while (lb->items) {
		temp = lb->items->next;
		mfe_item_destroy(lb->items);
		lb->items = temp;
	}
	lb->count = 0;
}

/*
 * Sorts the items in a list box according to the given comparison function.
 * The comparison function should behave like described in qsort(3).
 */
void mfe_list_box_sort(struct mfe_list_box *lb,
		       int (*cmp) (const void *, const void *))
{
	struct mfe_item *i1, *i1_p, *i2, *i2_p, *temp;

	i1 = i1_p = lb->items;
	if (!i1)
		return;

	while (i1->next) {
		i2 = i2_p = i1->next;
		while (i2) {
			if (cmp(i1->label->text, i2->label->text) > 0) {
				if (i1->next == i2) {
					/* adjacent items */

					i1->next = i2->next;
					i2->next = i1;

					temp = i1;
					i1 = i2;
					i2 = temp;

					if (i2 == lb->items) {
						/* head of the list */
						lb->items = i1;
						i1_p = i1;
					} else {
						/* other node */
						i1_p->next = i1;
					}
				} else {
					/* non-adjacent items */

					temp = i2->next;
					i2->next = i1->next;
					i1->next = temp;

					i2_p->next = i1;

					temp = i1;
					i1 = i2;
					i2 = temp;

					if (i2 == lb->items) {
						/* head of the list */
						lb->items = i1;
					} else {
						/* other node */
						i1_p->next = i1;
					}
				}
			}

			i2_p = i2;
			i2 = i2->next;
		}

		i1_p = i1;
		i1 = i1->next;
	}
}

/*
 * Adds a new item to the list box. The added item is the file or directory
 * corresponding to the path resulting from appending `base' to `name'.
 *
 */
int mfe_list_box_load_file_entry(struct mfe_list_box *lb, char *base,
				 char *name)
{
	char d_name[NAME_MAX + 2];
	struct mfe_item *item;
	char *path;

	item = mfe_item_create();
	if (!item) {
		return -1;
	}

	mfe_label_set_font(item->label, lb->font);
	mfe_label_remove_styles(item->label, MFE_LABEL_STYLE_SCROLL);

	/*
	 * Our directory entries will look like
	 * "directory/"
	 * and our files like
	 * "file"
	 */

	path = path_dup_base_and_name(base, name);
	if (!path) {
		mfe_item_destroy(item);
		return -1;
	}
	if (is_a_dir(path)) {
		strcpy(d_name, name);
		strcat(d_name, "/");
		mfe_item_set_text(item, d_name);
	} else {
		mfe_item_set_text(item, name);
	}
	free(path);

	mfe_list_box_link_item(lb, item);

	return 0;
}

/*
 * Reloads a list box with the contents of directory `path'.
 * If `preserve' is true, the old position in the file list is preserved if
 * possible.
 */
int mfe_list_box_reload_file_list(struct mfe_list_box *lb, char *path,
				  int preserve)
{
	struct mfe_item *item;
	struct dirent *entry;
	DIR *dir;
	char *s;
	int len;

	dir = opendir(path);
	if (!dir)
		return -1;

	mfe_list_box_destroy_all_items(lb);

	entry = readdir(dir);
	while (entry) {
		s = entry->d_name;
		len = strlen(s);
		if (!(len == 1 && s[0] == '.')) {
			if (mfe_list_box_load_file_entry(lb, path, s) == -1) {
				mfe_list_box_destroy_all_items(lb);
				closedir(dir);
				return -1;
			}
		}
		entry = readdir(dir);
	}
	closedir(dir);

	if (lb->count > 0) {
		mfe_list_box_sort(lb, path_compare);
		if (!preserve || lb->selected > lb->count ||
		    lb->first_shown > lb->selected) {
			lb->selected = 0;
			lb->first_shown = 0;
			item = lb->items;
		} else {
			item = mfe_list_box_get_item_by_index(lb, lb->selected);
		}
		mfe_item_select(item, 1);
	}

	return lb->count;
}

/*
 * Returns the width of a list box.
 */
static inline int mfe_list_box_get_width(struct mfe_list_box *lb)
{
	return lb->w;
}

/*
 * Returns the height of a list box.
 */
static inline int mfe_list_box_get_height(struct mfe_list_box *lb)
{
	return lb->h;
}

/*
 * Scrolls a list box in the given direction.
 */
static int mfe_list_box_scroll(struct mfe_list_box *lb, int direction)
{
	struct mfe_item *item;
	int h, index;
	int n = 0;

	if (!lb->items)
		return -1;

	item = mfe_list_box_get_selection(lb);
	if (!item) {
		item = lb->items;
		mfe_item_select(item, 1);
	}

	switch (direction) {
	case MFE_UP:
		if (lb->selected < lb->count - 1) {
			mfe_item_select(item, 0);
			lb->selected++;
			item = item->next;
			mfe_item_select(item, 1);

			index = lb->selected - lb->first_shown;
			h = lb->item_height * (index + 1);
			if (h > lb->h) {
				lb->first_shown++;
			}
			n = 1;
		}
		break;
	case MFE_DOWN:
		if (lb->selected > 0) {
			mfe_item_select(item, 0);
			lb->selected--;
			item = mfe_list_box_get_selection(lb);
			mfe_item_select(item, 1);

			if (lb->first_shown > lb->selected) {
				lb->first_shown = lb->selected;
			}
			n = 1;
		}
		break;
	}
	return n;
}

/*
 * Clips a list box into the given width and height.
 */
static void mfe_list_box_clip(struct mfe_list_box *lb, int w, int h)
{
	lb->max_w = w;
	lb->max_h = h;
}

/*
 * Calculates the rendered size of a list box.
 */
static void mfe_list_box_recalc_rendered_size(struct mfe_list_box *lb)
{
	struct mfe_item *item = lb->items;
	int w, h;

	w = 0;
	h = 0;

	while (item) {
		h += lb->item_height;
		if (h > lb->max_h) {
			break;
		}

		mfe_item_clip(item, lb->max_w, lb->item_height);
		mfe_item_recalc_rendered_size(item);

		w = mfe_item_get_width(item);
		if (w > lb->w)
			lb->w = w;

		item = item->next;
	}

	lb->w = min(lb->max_w, w);
	lb->h = min(lb->max_h, h);
}

/*
 * Starts a "fancy" transition for a list box.
 * This actually makes the list box unuseable, so the list box must be
 * destroyed after the fancy transition finishes.
 */
static void mfe_list_box_start_fancy_transition(struct mfe_list_box *lb)
{
	lb->saved_first_shown = lb->first_shown;
	lb->fancy = lb->first_shown;
}

/*
 * Checks if the "fancy" transition for a list box has finished.
 */
static int mfe_list_box_check_for_fancy_transition(struct mfe_list_box *lb)
{
	return lb->fancy >= 0;
}

/*
 * Performs a "fancy" transition step for a label.
 */
static void mfe_list_box_do_fancy_transition(struct mfe_list_box *lb,
					     int nr, int nr_text)
{
	struct mfe_item *item;

	if (nr_text == 0) {
		if (lb->selected > lb->first_shown) {
			lb->first_shown++;
			item = mfe_list_box_get_item_by_index(lb,
							      lb->selected + 1);
			while (item) {
				mfe_label_free_text(item->label);
				item = item->next;
			}
		} else {
			lb->first_shown = lb->saved_first_shown;
			lb->fancy = -1;
		}
	} else {
		if (lb->fancy >= 0) {
			item = mfe_list_box_get_item_by_index(lb, lb->fancy);
			if (item) {
				if (!mfe_item_is_selected(item)
				    && item->label->text) {
					mfe_label_set_text(item->label, NULL);
				}
				lb->fancy++;
			}
		}
	}
}

/*
 * Draws a list box to screen at the given rectangle.
 */
static void mfe_list_box_draw(SDL_Rect * dest, struct mfe_list_box *lb)
{
	struct mfe_label blank_label;
	struct mfe_item *item = lb->items;
	int y = 0, w;
	int nr = 0, nr_text = 0;
	int no_arrows = 0;
	SDL_Rect r;
	SDL_Surface *image;

	mfe_label_init(&blank_label);

	mfe_list_box_clip(lb, dest->w, dest->h);
	mfe_list_box_recalc_rendered_size(lb);

	w = dest->w;
	if (icon_arrow_up)
		w -= icon_arrow_up->w + 4;
	if (w < 0) {
		w = dest->w;
		no_arrows = 1;
	}

	item = mfe_list_box_get_item_by_index(lb, lb->first_shown);
	while (1) {
		r.x = dest->x;
		r.y = dest->y + y;
		r.w = w;
		r.h = lb->item_height;

		if (y + r.h > lb->h) {
			break;
		}

		if (item) {
			mfe_item_draw(&r, item);

			if (!mfe_item_is_selected(item) && item->label->text)
				nr_text++;
			nr++;

			item = item->next;
		} else {
			mfe_label_draw_blank(&r, &blank_label);
		}

		y += lb->item_height;
	}

	if (!no_arrows) {
		image = icon_arrow_up;
		r.x = dest->x + w + 4;
		r.y = dest->y;
		r.w = image->w;
		r.h = image->h;
		if (lb->first_shown > 0 && lb->fancy < 0) {
			mfe_image_draw(&r, image);
		} else {
			mfe_image_draw(&r, NULL);
		}

		image = icon_arrow_down;
		r.x = dest->x + w + 4;
		r.y = dest->y + dest->h - image->h;
		r.w = image->w;
		r.h = image->h;
		if (lb->count > lb->first_shown + nr && lb->fancy < 0) {
			mfe_image_draw(&r, image);
		} else {
			mfe_image_draw(&r, NULL);
		}
	}

	if (lb->fancy >= 0) {
		mfe_list_box_do_fancy_transition(lb, nr, nr_text);
	}
}

/*
 * Joystick "typematic" support.
 *
 */

/*
 * Initializes a joystick repeater with default values.
 */
static void mfe_joystick_repeater_init(struct mfe_joystick_repeater *jr,
				       SDL_Joystick * joystick)
{
	jr->joystick = joystick;
	jr->last_button_pressed = -1;
	jr->last_hat_used = -1;
	jr->last_state = 0;

	jr->delay = MFE_DIALOG_JOYSTICK_REPEAT_DELAY;
	jr->rate = MFE_DIALOG_JOYSTICK_REPEAT_RATE;
}

/*
 * Pushes the given event if the joystick repeat feature should trigger
 * at this precise moment.
 */
static int mfe_push_event_if_repeating(struct mfe_joystick_repeater *jr,
				       SDL_Event * event)
{
	Uint32 ticks;
	int retval = 0;

	ticks = SDL_GetTicks();

	if (!jr->ticks)
		jr->ticks = ticks;
	if (jr->repeating) {
		if (ticks > jr->ticks + (1000 / jr->rate)) {
			jr->ticks = ticks;
			SDL_PushEvent(event);
			retval = 1;

			jr->repeating++;
			if (jr->repeating <= 0)
				jr->repeating = 1;
		}
	} else {
		if (ticks > jr->ticks + jr->delay) {
			jr->ticks = ticks;
			retval = 1;

			jr->repeating = 1;
		}
	}
	return retval;
}

/*
 * Polls the joystick and generates repeating events as needed.
 */
static int mfe_poll_joystick(struct mfe_joystick_repeater *jr)
{
	SDL_Event event;
	int state;

	if (jr->last_button_pressed == -1 && jr->last_hat_used == -1) {
		jr->repeating = 0;
		jr->ticks = 0;
		return 0;
	}

	if (jr->last_button_pressed != -1) {
		state = SDL_JoystickGetButton(jr->joystick,
					      jr->last_button_pressed);

		event.jbutton.type = SDL_JOYBUTTONDOWN;
		event.jbutton.which = SDL_JoystickIndex(jr->joystick);
		event.jbutton.button = jr->last_button_pressed;
		event.jbutton.state = SDL_PRESSED;

		if (state && state == jr->last_state) {
			/* button pressed, and last time was pressed */
			mfe_push_event_if_repeating(jr, &event);
		} else if (state) {
			/* button pressed, and last time was not pressed */
			jr->last_state = state;
			jr->repeating = 0;
			jr->ticks = 0;
		} else {
			/* button not pressed */
			jr->last_button_pressed = -1;
		}
	}

	if (jr->last_hat_used != -1) {
		state = SDL_JoystickGetHat(jr->joystick, jr->last_hat_used);

		event.jhat.type = SDL_JOYHATMOTION;
		event.jhat.which = SDL_JoystickIndex(jr->joystick);
		event.jhat.hat = jr->last_hat_used;
		event.jhat.value = state;

		if (state && state == jr->last_state) {
			/* d-pad used, in same direction as last time */
			mfe_push_event_if_repeating(jr, &event);
		} else if (state) {
			/* d-pad used, in different direction */
			jr->last_state = state;
			jr->repeating = 0;
			jr->ticks = 0;
		} else {
			/* d-pad not used */
			jr->last_hat_used = -1;
		}
	}

	return 1;
}

/*
 * MPlayer slave mode driver.
 *
 *
 */

/*
 * Sends a slave command to MPlayer.
 */
static inline int mfe_send_command(struct mfe_session *s, char *cmd, ...)
{
	va_list ap;

	if (!s)
		return -1;

	va_start(ap, cmd);

	vfprintf(s->mplayer, cmd, ap);
	fprintf(s->mplayer, "\n");
	fflush(s->mplayer);

	va_end(ap);

	if (ferror(s->mplayer)) {
		s->flags |= MFE_SESSION_BROKEN;
		return -1;
	}
	return 0;
}

/*
 * Creates a MPlayer session.
 */
static struct mfe_session *mfe_create_session(char *file, int playlist,
					      char *extra_options)
{
	struct mfe_session *session;

	session = (struct mfe_session *)malloc(sizeof(*session));
	if (!session) {
		fprintf(stderr, "Can't allocate mplayer session.\n");
		return NULL;
	}

	session->cmd = vstrdup("mplayer -slave -noconsolecontrols -quiet"
			       " -nojoystick"
			       " -framedrop -cache 4096"
			       " %s %s '%s'",
			       extra_options,
			       playlist?"-playlist":"",
			       file);
	if (!session->cmd) {
		fprintf(stderr, "Can't allocate mplayer command line.\n");
		free(session);
		return NULL;
	}

	session->flags = 0;

	session->mplayer = popen(session->cmd, "w");
	if (!session->mplayer) {
		fprintf(stderr, "Can't lauch MPlayer: %s.\n", strerror(errno));
		free(session->cmd);
		free(session);
		return NULL;
	}

	return session;
}

/*
 * Destroys a MPlayer session.
 */
static void mfe_destroy_session(struct mfe_session *session)
{
	if (session) {
		alarm(MFE_CHILD_STUCK_TIMEOUT);
		pclose(session->mplayer);
		alarm(0);
		if (session->cmd)
			free(session->cmd);
		free(session);
	}
}

/*
 * Main part.
 *
 *
 */

static int mfe_flush_events;
struct mfe_joystick_repeater mfe_jrepeater;

static pid_t mfe_pid;
static struct mfe_session *mfe_session = NULL;

static struct mfe_file_dialog *mfe_file_dialog;
static SDL_Rect mfe_file_dialog_rect;

static char *mfe_dvd_device = MFE_DVD_DEVICE;
static char *mfe_dvd_mount_point = MFE_DVD_MOUNT_POINT;

static char *mfe_mplayer_extra_options = MFE_MPLAYER_EXTRA_OPTIONS;

static void mfe_file_dialog_reload(struct mfe_file_dialog *fd, int preserve);
static void mfe_directory_path_kick(struct mfe_file_dialog *fd, char *path);

/*
 * Sends a fake event to break SDL's wait event loop.
 */
static void wake_up_sdl(void)
{
	SDL_Event event;

	/* self wake up */
	event.type = SDL_USEREVENT;
	event.user.code = 0;
	event.user.data1 = NULL;
	event.user.data2 = NULL;
	SDL_PushEvent(&event);
}

/*
 * Our common signal handler.
 */
static void signal_handler(int sig)
{
	if (sig == SIGALRM) {
		killpg(mfe_pid, SIGTERM);
		mfe_flush_events = 1;
	}

	if (mfe_session) {
		mfe_session->flags |= MFE_SESSION_BROKEN;
	}

	wake_up_sdl();
}

/*
 * Displays a small text message at the bottom of the screen.
 */
static void mfe_display_bottom_message(char *text)
{
	struct mfe_label message, *l = &message;
	SDL_Rect message_rect, *r = &message_rect;
	int h;

	mfe_label_init(l);
	mfe_label_set_font(l, font_small);
	mfe_label_set_text(l, text);
	mfe_label_remove_styles(l, MFE_LABEL_STYLE_TRANSITION);

	h = TTF_FontHeight(l->font) + 4;
	r->x = MFE_HORZ_OVERSCAN;
	r->y = MFE_SCREEN_HEIGHT - h;
	r->w = MFE_SCREEN_WIDTH;
	r->h = h;

	mfe_label_draw(r, l);
}

/*
 * Mounts or dismounts DVD media, according to current mount status.
 */
static int mfe_toggle_mounted_media(struct mfe_file_dialog *fd)
{
	int retval;

	if (is_a_mount_point(mfe_dvd_mount_point)) {
		mfe_display_bottom_message("Unmounting media...");
		umount_iso9660(mfe_dvd_mount_point);
		retval = 1;
	} else {
		mfe_display_bottom_message("Mounting media... Please wait...");
		mount_iso9660(mfe_dvd_device, mfe_dvd_mount_point);
		retval = 0;
	}

	if (strncmp(fd->base, mfe_dvd_mount_point, strlen(mfe_dvd_mount_point))) {
		if (!retval)
			mfe_directory_path_kick(fd, mfe_dvd_mount_point);
		else
			mfe_file_dialog_reload(fd, 1);
	} else {
		mfe_directory_path_kick(fd, mfe_dvd_mount_point);
	}

	return retval;
}

/*
 * Starts playing the given media file.
 */
static int mfe_start_playing_file(char *file)
{
	struct mfe_joystick_repeater *jr = &mfe_jrepeater;
	int is_a_playlist = 0;

	mfe_display_bottom_message("Starting playback... May take a bit...");

	setenv("SDL_HACK_GCAIGM", "1", 1);

	if (ends_with(file, ".m3u") || ends_with(file, ".pls"))
		is_a_playlist = 1;

	mfe_session = mfe_create_session(file, is_a_playlist,
					 mfe_mplayer_extra_options);
	if (mfe_session) {
		jr->rate = MFE_PLAYER_JOYSTICK_REPEAT_RATE;
		jr->delay = MFE_PLAYER_JOYSTICK_REPEAT_DELAY;
		
		mfe_send_command(mfe_session, "osd_show_text www.gc-linux.org");
		return 0;
	}
	return -1;
}

/*
 * Stops playing the current played file.
 * It optionally repaints the background.
 */
static void mfe_stop_playing_file(int show_background)
{
	struct mfe_file_dialog *fd = mfe_file_dialog;
	struct mfe_joystick_repeater *jr = &mfe_jrepeater;

	if (mfe_session) {
		mfe_send_command(mfe_session, "quit");
		mfe_destroy_session(mfe_session);
		mfe_session = NULL;

		unsetenv("SDL_HACK_GCAIGM");

		if (show_background)
			mfe_draw_background(background);

		jr->rate = MFE_DIALOG_JOYSTICK_REPEAT_RATE;
		jr->delay = MFE_DIALOG_JOYSTICK_REPEAT_DELAY;
		
		mfe_file_dialog_reload(fd, 1);
	}
}

/*
 * Processes joystick events from the D-pad while playing media.
 */
static void mfe_process_joystick_hat_player(struct mfe_joystick_repeater *jr,
					    int value)
{
	if ((value & SDL_HAT_LEFT)) {
		mfe_send_command(mfe_session, "seek -5");
	}
	if ((value & SDL_HAT_RIGHT)) {
		mfe_send_command(mfe_session, "seek +5");
	}
}

/*
 * Processes joystick button events while playing media.
 */
static void mfe_process_joystick_button_player(struct mfe_joystick_repeater *jr,
					       int button)
{
	switch (button) {
	case BUTTON_A:
	case BUTTON_START:
		if (!jr->repeating) {
			if (!(mfe_session->flags & MFE_SESSION_PAUSED)) {
				mfe_send_command(mfe_session,
						 "osd_show_text paused");
			}
			mfe_send_command(mfe_session, "pause");
			mfe_session->flags ^= MFE_SESSION_PAUSED;
		}
		break;
	case BUTTON_B:
	case BUTTON_Z:
		mfe_stop_playing_file(1);
		break;
	case BUTTON_Y:
		/* free */
		break;
	case BUTTON_L:
		mfe_send_command(mfe_session, "seek -30");
		break;
	case BUTTON_R:
		mfe_send_command(mfe_session, "seek +30");
		break;
	case BUTTON_X:
		if (!jr->repeating)
			mfe_send_command(mfe_session, "sub_select");
		break;
	default:
		break;
	}
}

/*
 * File dialog functions.
 *
 *
 */

/*
 * Returns in `pathp' a full pathname built by appending the text from the
 * current selected entry in the given list box to `base'.
 */
char *mfe_build_full_name_of_selected_entry(struct mfe_list_box *lb,
					    char *base, char **pathp)
{
	struct mfe_item *item;
	struct mfe_label *l;
	char *entry;
	int len;

	item = mfe_list_box_get_selection(lb);
	if (item && item->label) {
		l = item->label;
		entry = l->text;
		if (entry) {
			if (*pathp)
				free(*pathp);
			*pathp =
			    (char *)malloc(strlen(base) + strlen(entry) + 1);
			if (*pathp) {
				len = strlen(base);
				strcpy(*pathp, base);
				if (base[len - 1] != '/')
					strcat(*pathp, "/");
				strcat(*pathp, entry);
				if (l->added_separator) {
					len = strlen(*pathp);
					(*pathp)[len - sizeof(MFE_SEPARATOR) +
						 1] = 0;
				}
			}
		}
	}
	return *pathp;
}

/*
 * Initializes a file dialog with default values.
 */
static void mfe_file_dialog_init(struct mfe_file_dialog *fd)
{
	SDL_Color white = { 255, 255, 255, 0 };

	mfe_label_add_styles(fd->l, MFE_LABEL_STYLE_OSCILATE |
			     MFE_LABEL_STYLE_BOX_COMPLETE);
	mfe_label_set_color_bg(fd->l, white);
	mfe_label_set_text(fd->l, fd->base);

	fd->path = NULL;
	fd->action = MFE_ACTION_NONE;

	if (fd->base)
		mfe_list_box_reload_file_list(fd->lb, fd->base, 0);
}

/*
 * Creates and initializes a file dialog.
 */
struct mfe_file_dialog *mfe_file_dialog_create(char *base)
{
	struct mfe_label *l;
	struct mfe_list_box *lb;
	struct mfe_file_dialog *fd;

	fd = (struct mfe_file_dialog *)malloc(sizeof(*fd));
	if (!fd)
		return NULL;

	fd->base = strdup(base);
	if (!fd->base) {
		free(fd);
		return NULL;
	}

	l = mfe_label_create();
	if (!l) {
		free(fd->base);
		free(fd);
		return NULL;
	}
	fd->l = l;

	lb = mfe_list_box_create();
	if (!lb) {
		mfe_label_destroy(l);
		free(fd->base);
		free(fd);
		return NULL;
	}
	fd->lb = lb;

	mfe_file_dialog_init(fd);

	return fd;
}

/*
 * Starts a "fancy" transition effect for the file dialog.
 */
static void mfe_file_dialog_start_fancy_transition(struct mfe_file_dialog *fd)
{
	mfe_list_box_start_fancy_transition(fd->lb);
	mfe_label_start_fancy_transition(fd->l);
}

/*
 * Checks if a "fancy" transition has finished for the given file dialog.
 */
static int mfe_file_dialog_check_for_fancy_transition(struct mfe_file_dialog
						      *fd)
{
	return mfe_list_box_check_for_fancy_transition(fd->lb) ||
	    mfe_label_check_for_fancy_transition(fd->l);
}

/*
 * Draws a file dialog to screen at the given rectangle.
 */
static void mfe_file_dialog_draw(SDL_Rect * dest, struct mfe_file_dialog *fd)
{
	int h, font_height;
	SDL_Rect r;

	font_height = TTF_FontHeight(fd->l->font);

	h = font_height + 2 * AMPL_Y;
	if (h > dest->h)
		return;

	r.x = dest->x;
	r.y = dest->y + dest->h - h;
	r.w = dest->w;
	r.h = h;
	mfe_label_draw(&r, fd->l);

	r.x = dest->x;
	r.y = dest->y;
	r.w = dest->w;
	r.h = dest->h - h;
	mfe_list_box_draw(&r, fd->lb);
}

/*
 * Reloads the contents of a file dialog.
 */
static void mfe_file_dialog_reload(struct mfe_file_dialog *fd, int preserve)
{
	mfe_list_box_reload_file_list(fd->lb, fd->base, preserve);
	mfe_label_set_text(fd->l, fd->base);

	mfe_draw_background(background);
}

/*
 * Goes back one directory level in a file dialog.
 */
static void mfe_file_dialog_entry_back(struct mfe_file_dialog *fd)
{
	path_remove_last_name(fd->base);

	mfe_file_dialog_reload(fd, 0);

}

/*
 * Goes up one position in a file dialog.
 */
static void mfe_file_dialog_entry_up(struct mfe_file_dialog *fd)
{
	if (mfe_list_box_scroll(fd->lb, MFE_UP) > 0) {
		mfe_build_full_name_of_selected_entry(fd->lb, fd->base,
						      &fd->path);
	}
}

/*
 * Goes down one position in a file dialog.
 */
static void mfe_file_dialog_entry_down(struct mfe_file_dialog *fd)
{
	if (mfe_list_box_scroll(fd->lb, MFE_DOWN) > 0) {
		mfe_build_full_name_of_selected_entry(fd->lb, fd->base,
						      &fd->path);
	}
}

/*
 * "Kicks" (launches) the given path using the specified file dialog.
 */
static void mfe_directory_path_kick(struct mfe_file_dialog *fd, char *path)
{
	mode_t mode;

	mfe_label_set_text(fd->l, NULL);

	if (fd->path)
		free(fd->path);
	fd->path = strdup(path);

	mode = get_file_type(fd->path);
	if (mode == (mode_t) - 1)
		return;

	if (is_a_dir(fd->path)) {
		if (fd->base)
			free(fd->base);
		fd->base = realpath(fd->path, NULL);

		mfe_file_dialog_reload(fd, 0);
	} else {
		fd->action = MFE_ACTION_PLAY;
		mfe_file_dialog_start_fancy_transition(fd);
	}
}

/*
 * "Kicks" the selected entry of a file dialog.
 */
static void mfe_directory_entry_kick(struct mfe_file_dialog *fd)
{
	char *path = NULL;

	mfe_build_full_name_of_selected_entry(fd->lb, fd->base, &path);
	mfe_directory_path_kick(fd, path);
	if (path)
		free(path);

}

/*
 * Processes joystick events from the D-pad while in the file dialog.
 */
static void mfe_process_joystick_hat_menu(struct mfe_file_dialog *fd,
					  struct mfe_joystick_repeater *jr,
					  int hat, int value)
{
	if ((value & SDL_HAT_UP)) {
		mfe_file_dialog_entry_up(fd);
	}
	if ((value & SDL_HAT_DOWN)) {
		mfe_file_dialog_entry_down(fd);
	}
}

/*
 * Processes joystick button events while in the file dialog.
 */
static void mfe_process_joystick_button_menu(struct mfe_file_dialog *fd,
					     struct mfe_joystick_repeater *jr,
					     int button)
{
	switch (button) {
	case BUTTON_A:
	case BUTTON_START:
		if (!jr->repeating)
			mfe_directory_entry_kick(fd);
		break;
	case BUTTON_B:
		if (!jr->repeating)
			mfe_file_dialog_entry_back(fd);
		break;
	case BUTTON_X:
		mfe_file_dialog_entry_up(fd);
		break;
	case BUTTON_Y:
		mfe_file_dialog_entry_down(fd);
		break;
	case BUTTON_L:
	case BUTTON_R:
		/* mount/unmount media if pressed for more than 1 second */
		if (jr->repeating > 1 * jr->rate) {
			fd->action = MFE_ACTION_TOGGLE_MOUNTED_MEDIA;
			mfe_label_start_fancy_transition(fd->l);
			/* actively prevent repeating */
			jr->last_button_pressed = -1;
		}
		break;
	case BUTTON_Z:
		/* quit if pressed for more than 2 seconds */
		if (jr->repeating >= 2 * jr->rate)
			exit(1);
		break;
	default:
		break;
	}
}

/*
 * Processes events.
 */
static void mfe_process_event(SDL_Event * event)
{
	struct mfe_file_dialog *fd = mfe_file_dialog;
	struct mfe_joystick_repeater *jr = &mfe_jrepeater;

	switch (event->type) {
	case SDL_JOYBUTTONDOWN:
		jr->last_button_pressed = event->jbutton.button;
		if (mfe_session) {
			mfe_process_joystick_button_player(jr, event->jbutton.
							   button);
		} else {
			mfe_process_joystick_button_menu(fd, jr,
							 event->jbutton.button);
		}
		break;
	case SDL_JOYHATMOTION:
		jr->last_hat_used = event->jhat.hat;
		jr->last_state = event->jhat.value;
		if (mfe_session) {
			mfe_process_joystick_hat_player(jr, event->jhat.value);
		} else {
			mfe_process_joystick_hat_menu(fd, jr,
						      event->jhat.hat,
						      event->jhat.value);
		}
		break;
	case SDL_KEYDOWN:
		if (event->key.keysym.sym == SDLK_ESCAPE)
			exit(1);
		break;
	case SDL_QUIT:
		return;
	default:
		break;
	}
}

/*
 * Dispatches the current scheduled action for a file dialog.
 */
static int mfe_action_dispatcher(struct mfe_file_dialog *fd)
{
	switch (fd->action) {
	case MFE_ACTION_PLAY:
		if (!mfe_file_dialog_check_for_fancy_transition(fd)) {
			fd->action = MFE_ACTION_NONE;
			mfe_start_playing_file(fd->path);
		}
		break;
	case MFE_ACTION_TOGGLE_MOUNTED_MEDIA:
		if (!mfe_label_check_for_fancy_transition(fd->l)) {
			fd->action = MFE_ACTION_NONE;
			mfe_toggle_mounted_media(fd);
		}
		break;
	}
	return 0;
}

/*
 * Dispatches events.
 */
static void mfe_event_dispatcher(void)
{
	struct mfe_file_dialog *fd = mfe_file_dialog;
	SDL_Event event;

	if (fd->action) {
		mfe_action_dispatcher(fd);
	} else {
		while (SDL_PollEvent(&event)) {
			if (!mfe_flush_events)
				mfe_process_event(&event);
		}
		mfe_flush_events = 0;

		if (!mfe_poll_joystick(&mfe_jrepeater)) {
			if (mfe_session) {
				SDL_WaitEvent(NULL);
			}
		}
	}
}

#define MFE_RENDER_PERIOD (1000/75)
#define MFE_WATCHDOG_PERIOD 1000

/*
 * Renders any visual objects to screen.
 */
static void mfe_render(void)
{
	static Uint32 old_ticks = 0;
	Uint32 ticks;
	struct mfe_file_dialog *fd;
	SDL_Rect *r;

	ticks = SDL_GetTicks();

	if (ticks > old_ticks + MFE_RENDER_PERIOD) {
		old_ticks = ticks;

		if (!mfe_session) {
			fd = mfe_file_dialog;
			r = &mfe_file_dialog_rect;
			mfe_file_dialog_draw(r, fd);
		}
	}
}

/*
 * Takes care of things that may go wrong.
 */
static void mfe_watchdog(void)
{
	static Uint32 old_ticks = 0;
	Uint32 ticks = SDL_GetTicks();

	if (mfe_session) {
		if (ticks > old_ticks + MFE_WATCHDOG_PERIOD) {
			mfe_send_command(mfe_session, "nop");
			old_ticks = ticks;
		}

		if ((mfe_session->flags & MFE_SESSION_BROKEN)) {
			mfe_stop_playing_file(1);
		}
	}
}

static const char *__progname;

/*
 * Sets the DVD device file name to use.
 */
static int set_dvd_device(char *device)
{
	mode_t mode;

	mode = get_file_type(device);
	if (mode == (mode_t) - 1) {
		fprintf(stderr, "Unable to determine the type of %s: %s.\n",
			device, strerror(errno));
		return -1;
	}

	if (S_ISBLK(mode)) {
		mfe_dvd_device = device;
	} else {
		fprintf(stderr, "%s is not a valid block device.\n", device);
		return -1;
	}

	return 0;
}

/*
 * Sets the DVD mount point to use.
 */
static int set_dvd_mount_point(char *path)
{
	if (!is_a_dir(path)) {
		fprintf(stderr, "Sorry, %s is not a valid directory.\n", path);
		return -1;
	}

	mfe_dvd_mount_point = path;

	return 0;
}

/*
 * Displays current version of the "mfe".
 */
void version(void)
{
	fprintf(stderr, "version %s\n", MFE_VERSION);
	exit(2);
}

/*
 * Displays usage instructions for "mfe".
 */
void usage(void)
{
	fprintf(stderr,
		"Usage: %s [OPTIONS] [DIRNAME]" "\n"
		"  -z, --dvd-device=DEVICE              "
		" use DEVICE dvd device file" "\n"
		"\t" "(default is " MFE_DVD_DEVICE ")" "\n"
		"  -d, --dvd-mount-point=DIR            "
		" mount dvd media at DIR" "\n"
		"\t" "(default is " MFE_DVD_MOUNT_POINT ")" "\n"
		"  -m, --mount                          "
		" try to mount media during startup" "\n"
		"  -u, --umount                         "
		" try to unmount media during startup" "\n"
		"  -x, --extra-mplayer-options=OPTIONS  "
		" pass OPTIONS to MPlayer" "\n"
		"\t" "(default is \"" MFE_MPLAYER_EXTRA_OPTIONS "\")" "\n"
		"  -h, --help                           "
		" call for help" "\n"
		"  -v, --version                        "
		" display version" "\n", __progname);
	exit(1);
}

int main(int argc, char *argv[])
{
	struct mfe_file_dialog *fd;
	SDL_Rect *r;

	char *p, *base;
	int ch;

	int try_to_mount = 0, try_to_umount = 0;

	p = strrchr(argv[0], '/');
	__progname = (p && p[1]) ? p + 1 : argv[0];

	/* several signals we care about */
	signal(SIGPIPE, signal_handler);
	signal(SIGCHLD, signal_handler);
	signal(SIGALRM, signal_handler);

	/* prepare things in case we need to kill our childs */
	mfe_pid = getpid();
	setpgid(mfe_pid, mfe_pid);

	/*
	 * Process command line options.
	 */

	struct option long_options[] = {
		{"dvd-device", 1, NULL, 'z'},
		{"dvd-mount-point", 1, NULL, 'd'},
		{"mount", 0, NULL, 'm'},
		{"umount", 0, NULL, 'u'},
		{"extra-mplayer-options", 1, NULL, 'x'},
		{"version", 0, NULL, 'v'},
		{"help", 0, NULL, 'h'},
		{0, 0, 0, 0}
	};
#define SHORT_OPTIONS "z:d:mux:vh"

	while ((ch = getopt_long(argc, argv, SHORT_OPTIONS,
				 long_options, NULL)) != -1) {
		switch (ch) {
		case 'z':
			if (set_dvd_device(optarg) < 0)
				exit(2);
			break;
		case 'd':
			if (set_dvd_mount_point(optarg) < 0)
				exit(3);
			break;
		case 'm':
			try_to_mount = 1;
			try_to_umount = 0;
			break;
		case 'u':
			try_to_mount = 0;
			try_to_umount = 1;
			break;
		case 'x':
			mfe_mplayer_extra_options = optarg;
			break;
		case 'v':
			version();
			break;
		case 'h':
		case '?':
		default:
			usage();
			break;
		}
	}

	if (argc - optind != 1) {
		usage();
	}

	/* we start our file dialog from an existing directory */
	base = realpath(argv[optind], NULL);
	if (!is_a_dir(base)) {
		fprintf(stderr, "Sorry, %s is not a valid directory.\n", base);
		exit(1);
	}

	/*
	 * Start the graphical user interface. We need to do that
	 * before creating any graphical user interface object.
	 */
	mfe_init_gui(SDL_FULLSCREEN);
	mfe_draw_background(background);

	/*
	 * The file dialog becomes the central part of the GUI.
	 */
	fd = mfe_file_dialog_create(base);
	if (!fd) {
		fprintf(stderr, "Unable to create the file dialog.\n");
		exit(4);
	}
	free(base);
	mfe_file_dialog = fd;
	mfe_file_dialog_init(fd);

	r = &mfe_file_dialog_rect;
	r->x = MFE_HORZ_OVERSCAN;
	r->y = MFE_VERT_OVERSCAN + fd->lb->item_height;
	r->w = MFE_SCREEN_WIDTH;
	r->h = MFE_SCREEN_HEIGHT - 2 * fd->lb->item_height;

	/* do not forget the joystick... */
	mfe_joystick_repeater_init(&mfe_jrepeater, joystick);

	if (try_to_mount) {
		if (!is_a_mount_point(mfe_dvd_mount_point)) {
			fd->action = MFE_ACTION_TOGGLE_MOUNTED_MEDIA;
			mfe_label_start_fancy_transition(fd->l);
		}
	}

	if (try_to_umount) {
		if (is_a_mount_point(mfe_dvd_mount_point)) {
			fd->action = MFE_ACTION_TOGGLE_MOUNTED_MEDIA;
			mfe_label_start_fancy_transition(fd->l);
		}
	}

	/* ok, here we go */
	while (1) {
		mfe_watchdog();
		mfe_render();
		mfe_event_dispatcher();
	}

	exit(0);
}
